"""
This file is imported with some modification from gdown repo, many thanks and support to them.
gdown repo: https://github.com/wkentaro/gdown
"""
from __future__ import print_function

import glob
import json
import os
import os.path as osp
import re
import shutil
import sys
import tempfile
import textwrap
import time

import requests
import six
import tqdm

from .parse_url import parse_url


CHUNK_SIZE = 512 * 1024  # 512KB
home = osp.expanduser("~")


if hasattr(textwrap, "indent"):
    indent_func = textwrap.indent
else:

    def indent_func(text, prefix):
        def prefixed_lines():
            for line in text.splitlines(True):
                yield (prefix + line if line.strip() else line)

        return "".join(prefixed_lines())


def get_url_from_gdrive_confirmation(contents):
    url = ""
    for line in contents.splitlines():
        m = re.search(r'href="(\/uc\?export=download[^"]+)', line)
        if m:
            url = "https://docs.google.com" + m.groups()[0]
            url = url.replace("&amp;", "&")
            break
        m = re.search("confirm=([^;&]+)", line)
        if m:
            confirm = m.groups()[0]
            url = re.sub(
                r"confirm=([^;&]+)", r"confirm={}".format(confirm), url
            )
            break
        m = re.search('"downloadUrl":"([^"]+)', line)
        if m:
            url = m.groups()[0]
            url = url.replace("\\u003d", "=")
            url = url.replace("\\u0026", "&")
            break
        m = re.search('<p class="uc-error-subcaption">(.*)</p>', line)
        if m:
            error = m.groups()[0]
            raise RuntimeError(error)
    if not url:
        raise RuntimeError(
            "Cannot retrieve the public link of the file. "
            "You may need to change the permission to "
            "'Anyone with the link', or have had many accesses."
        )
    return url


def download(
    url=None,
    output=None,
    quiet=False,
    proxy=None,
    speed=None,
    use_cookies=True,
    verify=True,
    id=None,
    fuzzy=False,
    resume=False,
):
    """Download file from URL.

    Parameters
    ----------
    url: str
        URL. Google Drive URL is also supported.
    output: str
        Output filename. Default is basename of URL.
    quiet: bool
        Suppress terminal output. Default is False.
    proxy: str
        Proxy.
    speed: float
        Download byte size per second (e.g., 256KB/s = 256 * 1024).
    use_cookies: bool
        Flag to use cookies. Default is True.
    verify: bool or string
        Either a bool, in which case it controls whether the server’s TLS
        certificate is verified, or a string, in which case it must be a path
        to a CA bundle to use. Default is True.
    id: str
        Google Drive's file ID.
    fuzzy: bool
        Fuzzy extraction of Google Drive's file Id. Default is False.
    resume: bool
        Resume the download from existing tmp file if possible.
        Default is False.

    Returns
    -------
    output: str
        Output filename.
    """
    if not (id is None) ^ (url is None):
        raise ValueError("Either url or id has to be specified")
    if id is not None:
        url = "https://drive.google.com/uc?id={id}".format(id=id)

    url_origin = url
    sess = requests.session()
    res = None
    # Load cookies
    cache_dir = osp.join(home, ".cache", "gdown")
    if not osp.exists(cache_dir):
        os.makedirs(cache_dir)
    cookies_file = osp.join(cache_dir, "cookies.json")
    if osp.exists(cookies_file) and use_cookies:
        with open(cookies_file) as f:
            cookies = json.load(f)
        for k, v in cookies:
            sess.cookies[k] = v

    if proxy is not None:
        sess.proxies = {"http": proxy, "https": proxy}
        print("Using proxy:" + proxy)

    gdrive_file_id, is_gdrive_download_link = parse_url(url, warning=not fuzzy)

    if fuzzy and gdrive_file_id:
        # overwrite the url with fuzzy match of a file id
        url = "https://drive.google.com/uc?id={id}".format(id=gdrive_file_id)
        url_origin = url
        is_gdrive_download_link = True

    headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"  # NOQA
    }

    while True:
        try:
            if id is None:
                res = sess.get(url,headers=headers, stream=True, timeout= (20,5))
            else:
                res = sess.get(url, headers=headers, stream=True, verify=verify, timeout = 30)

        except requests.exceptions.ProxyError as e:
            print("An error has occurred using proxy")
            return "Error"
        except requests.exceptions.ConnectionError as e:
            print("An error has occurred connectiong to download server")
            return "Error"
        except requests.exceptions.ReadTimeout as e:
            print("An error has occurred connectiong to download server")
            return "Error"

        

        # Save cookies
        with open(cookies_file, "w") as f:
            cookies = [
                (k, v)
                for k, v in sess.cookies.items()
                if not k.startswith("download_warning_")
            ]
            json.dump(cookies, f, indent=2)

        if "Content-Disposition" in res.headers:
            # This is the file
            break
        if not (gdrive_file_id and is_gdrive_download_link):
            break

        # Need to redirect with confirmation
        try:
            url = get_url_from_gdrive_confirmation(res.text)
        except RuntimeError as e:
            print("Access denied with the following error:")
            error = "\n".join(textwrap.wrap(str(e)))
            error = indent_func(error, "\t")
            print("\n"+ error+"\n")
            print("You may still be able to access the file from the browser")
            print("\n\t"+url_origin+"\n")
            return "Error"

    if gdrive_file_id and is_gdrive_download_link:
        m = re.search('filename="(.*)"', res.headers["Content-Disposition"])
        filename_from_url = m.groups()[0]
    else:
        filename_from_url = osp.basename(url)

    if output is None:
        output = filename_from_url

    output_is_path = isinstance(output, six.string_types)
    if output_is_path and output.endswith(osp.sep):
        if not osp.exists(output):
            os.makedirs(output)
        output = osp.join(output, filename_from_url)

    if output_is_path:
        existing_tmp_files = glob.glob("{}*".format(output))
        if resume and existing_tmp_files:
            tmp_file = existing_tmp_files[0]
        else:
            resume = False
            tmp_file = tempfile.mktemp(
                suffix=tempfile.template,
                prefix=osp.basename(output),
                dir=osp.dirname(output),
            )
            tmp_file = osp.join(osp.dirname(output),osp.basename(output))
            print(tmp_file)
        f = open(tmp_file, "wb")
    else:
        tmp_file = None
        f = output
    print(f.tell())
    if tmp_file is not None and f.tell() != 0:
        headers["Range"] = "bytes={}-".format(f.tell())
        try:
            if id is None:
                res = sess.get(url, stream=True, timeout = (20,5))
            else:
                res = sess.get(url, headers=headers, stream=True, verify=verify, timeout = 30)

        except requests.exceptions.ConnectionError as e:
            print("An error has occurred connectiong to download server")
            return "Error"
        except requests.exceptions.ReadTimeout as e:
            print("An error has occurred connectiong to download server")
            return "Error"    

    if not quiet:
        print("Downloading started ...")
        if resume:
            print("Resume: ", tmp_file)
        print("From: "+ url_origin)
        print(
            "To: "+ osp.abspath(output) if output_is_path else output,)

    try:
        total = res.headers.get("Content-Length")
        if total is not None:
            total = int(total)
        if not quiet:
            pbar = tqdm.tqdm(total=total, unit="B", unit_scale=True)
        t_start = time.time()
        for chunk in res.iter_content(chunk_size=CHUNK_SIZE):
            if time.time() - t_start > 50:
                return "Error"
            f.write(chunk)
            if not quiet:
                pbar.update(len(chunk))
            if speed is not None:
                elapsed_time_expected = 1.0 * pbar.n / speed
                
                elapsed_time = time.time() - t_start
                if elapsed_time < elapsed_time_expected:
                    time.sleep(elapsed_time_expected - elapsed_time)
            t_start = time.time()
            
                    
        if not quiet:
            pbar.close()
        if tmp_file:
            f.close()
            shutil.move(tmp_file, output)
    except IOError as e:
        print(e)
        return "Error"
    except requests.exceptions.RequestException as e:  # This is the correct syntax
            return "Error"
    
    except requests.exceptions.ConnectionError as e:
        print("An error has occurred connectiong to download server")
        return "Error"
    except requests.exceptions.ReadTimeout as e:
        print("An error has occurred connectiong to download server")
        return "Error"
    finally:
        sess.close()

    return output



"""
This code is built by open source github community many thanks to them
"""
import math

class AnchorsOptions:
	def __init__(self, input_size_width, input_size_height, min_scale, max_scale
		, num_layers, feature_map_width, feature_map_height
		, strides, aspect_ratios, anchor_offset_x=0.5, anchor_offset_y=0.5
		, reduce_boxes_in_lowest_layer=False, interpolated_scale_aspect_ratio=1.0
		, fixed_anchor_size=False):
		# Size of input images.
		self.input_size_width = input_size_width
		self.input_size_height = input_size_height
		# Min and max scales for generating anchor boxes on feature maps.
		self.min_scale = min_scale
		self.max_scale = max_scale
		# The offset for the center of anchors. The value is in the scale of stride.
		# E.g. 0.5 meaning 0.5 * |current_stride| in pixels.
		self.anchor_offset_x = anchor_offset_x
		self.anchor_offset_y = anchor_offset_y
		# Number of output feature maps to generate the anchors on.
		self.num_layers = num_layers
		# Sizes of output feature maps to create anchors. Either feature_map size or
		# stride should be provided.
		self.feature_map_width = feature_map_width
		self.feature_map_height = feature_map_height
		self.feature_map_width_size = len(feature_map_width)
		self.feature_map_height_size = len(feature_map_height)
		# Strides of each output feature maps.
		self.strides = strides
		self.strides_size = len(strides)
		# List of different aspect ratio to generate anchors.
		self.aspect_ratios = aspect_ratios
		self.aspect_ratios_size = len(aspect_ratios)
		# A boolean to indicate whether the fixed 3 boxes per location is used in the lowest layer.
		self.reduce_boxes_in_lowest_layer = reduce_boxes_in_lowest_layer
		# An additional anchor is added with this aspect ratio and a scale
		# interpolated between the scale for a layer and the scale for the next layer
		# (1.0 for the last layer). This anchor is not included if this value is 0.
		self.interpolated_scale_aspect_ratio = interpolated_scale_aspect_ratio
		# Whether use fixed width and height (e.g. both 1.0f) for each anchor.
		# This option can be used when the predicted anchor width and height are in  pixels.
		self.fixed_anchor_size = fixed_anchor_size
	def to_string(self):
		return 'input_size_width: {:}\ninput_size_height: {:}\nmin_scale: {:}\nmax_scale: {:}\nanchor_offset_x: {:}\nanchor_offset_y: {:}\nnum_layers: {:}\nfeature_map_width: {:}\nfeature_map_height: {:}\nstrides: {:}\naspect_ratios: {:}\nreduce_boxes_in_lowest_layer: {:}\ninterpolated_scale_aspect_ratio: {:}\nfixed_anchor_size: {:}'\
		.format(self.input_size_width, self.input_size_height, self.min_scale, self.max_scale
			, self.anchor_offset_x, self.anchor_offset_y, self.num_layers
			, self.feature_map_width, self.feature_map_height, self.strides, self.aspect_ratios
			, self.reduce_boxes_in_lowest_layer, self.interpolated_scale_aspect_ratio
			, self.fixed_anchor_size)

class Anchor:
	def __init__(self, x_center, y_center, h, w):
		self.x_center = x_center
		self.y_center = y_center
		self.h = h
		self.w = w
	def to_string(self):
		return 'x_center: {:}, y_center: {:}, h: {:}, w: {:}'.format(self.x_center, self.y_center, self.h, self.w)

def gen_anchors(options):
	anchors = []
	# Verify the options.
	if (options.strides_size != options.num_layers):
		print("strides_size and num_layers must be equal.")
		return []

	layer_id = 0
	while (layer_id < options.strides_size):
		anchor_height = []
		anchor_width = []
		aspect_ratios = []
		scales = []

		# For same strides, we merge the anchors in the same order.
		last_same_stride_layer = layer_id
		while (last_same_stride_layer < options.strides_size and options.strides[last_same_stride_layer] == options.strides[layer_id]):
			scale = options.min_scale + (options.max_scale - options.min_scale) * 1.0 * last_same_stride_layer / (options.strides_size - 1.0)
			if (last_same_stride_layer == 0 and options.reduce_boxes_in_lowest_layer):
				# For first layer, it can be specified to use predefined anchors.
				aspect_ratios.append(1.0)
				aspect_ratios.append(2.0)
				aspect_ratios.append(0.5)
				scales.append(0.1)
				scales.append(scale)
				scales.append(scale)
			else:
				for aspect_ratio_id in range(options.aspect_ratios_size):
					aspect_ratios.append(options.aspect_ratios[aspect_ratio_id])
					scales.append(scale)
				
				if (options.interpolated_scale_aspect_ratio > 0.0):
					scale_next = 1.0 if last_same_stride_layer == options.strides_size - 1 else options.min_scale + (options.max_scale - options.min_scale) * 1.0 * (last_same_stride_layer+1) / (options.strides_size - 1.0)
					scales.append(math.sqrt(scale * scale_next))
					aspect_ratios.append(options.interpolated_scale_aspect_ratio)
			last_same_stride_layer += 1
		for i in range(len(aspect_ratios)):
			ratio_sqrts = math.sqrt(aspect_ratios[i])
			anchor_height.append(scales[i] / ratio_sqrts)
			anchor_width.append(scales[i] * ratio_sqrts)

		feature_map_height = 0
		feature_map_width = 0
		if (options.feature_map_height_size > 0):
			feature_map_height = options.feature_map_height[layer_id]
			feature_map_width = options.feature_map_width[layer_id]
		else:
			stride = options.strides[layer_id]
			feature_map_height = math.ceil(1.0 * options.input_size_height / stride)
			feature_map_width = math.ceil(1.0 * options.input_size_width / stride)

		for y in range(feature_map_height):
			for x in range(feature_map_width):
				for anchor_id in range(len(anchor_height)):
					# TODO: Support specifying anchor_offset_x, anchor_offset_y.
					x_center = (x + options.anchor_offset_x) * 1.0 / feature_map_width
					y_center = (y + options.anchor_offset_y) * 1.0 / feature_map_height
					w = 0
					h = 0
					if (options.fixed_anchor_size):
						w = 1.0
						h = 1.0
					else:
						w = anchor_width[anchor_id]
						h = anchor_height[anchor_id]
					new_anchor = Anchor(x_center, y_center, h, w)
					anchors.append(new_anchor)
		layer_id = last_same_stride_layer
	return anchors


import cv2
import numpy as np
import os
from .download import download
from .BlazeDetector import BlazeFaceDetector


class face_lib:

    def __init__(self, detection_model = "front"):
        """
        input: [detection_model] a string holds the detection model name
                there are two choices "front" which is fast, "back" which is a little bit slower but more accurate
                for most use cases "front" will be just fine
        """
        BASE_DIR = os.path.dirname(__file__)

        self.__faceDetector = BlazeFaceDetector(typ=detection_model)
        #self.profileClassfier = cv2.CascadeClassifier("haarcascade_profileface.xml")
        self.__faceEmbeddingNet = None
        file_exists = False
        try:
            file_exists = os.path.exists(os.path.join(BASE_DIR , "graph_final.pb"))
            self.__faceEmbeddingNet = cv2.dnn.readNetFromTensorflow(os.path.join(BASE_DIR , "graph_final.pb"))
        except:
            ############## Download Face Recognition model fom github #############
            if file_exists :
                print("Face recognition model is corrupted downloading it again, please wait ...")
                print("This download will be done once if no errors happened, don't worry ...")
            else:
                print("Downloading face recognition model for the first time ...")
                print("This download will be done once if no errors happened, don't worry ...")
            
            print("If the download didn't done correctly you can create the instance again it will try to download it again automatically for you")
            
            check_error = download("https://github.com/a-akram-98/face_lib/releases/download/v1.0.5/graph_final.pb", os.path.join(BASE_DIR , "graph_final.pb"), quiet=False)
            if check_error == "Error":
                print("It seems github release is too slow, switching to the download from Google Drive")
                check_error = download(id = "1pNU-V31cdgCSRu9XQqoSv1tjbuAfVOSG",output= os.path.join(BASE_DIR , "graph_final.pb"), quiet=False)
                if check_error == "Error":
                    print("Download Failed, It seems your connection is slow ...")
                    print("You can download the model manually from this link: "+ "https://github.com/a-akram-98/face_lib/releases/download/v1.0.5/graph_final.pb")
                    print("then add it to the following path:")
                    print(BASE_DIR)
                    return


            self.__faceEmbeddingNet = cv2.dnn.readNetFromTensorflow(os.path.join(BASE_DIR , "graph_final.pb"))
            
    def set_detection_params(self, scoreThreshold=0.7, iouThreshold=0.3):
        """
        inputs: [scoreThreshold] an float between 0-1 represents the confidence level thershold
                [iouThreshold] an float between 0-1 represents the intersection over union threshold
        """
        self.__faceDetector.scoreThreshold = scoreThreshold
        self.__faceDetector.sigmoidScoreThreshold = np.log(scoreThreshold/(1-scoreThreshold))
        self.__faceDetector.iouThreshold = iouThreshold

    
    def recognition_pipeline(self, face_img, gt_img, only_face_gt = False, threshold = 0.92):
        """
        input: test image (face_img), and ground truth image (given from cv.imread(), expecting BGR image),
        return: [True] if the person in the test image is the same as frounf truth image,
                [False] if there is no person detected or the preson in test not the one in the ground truth
                [no_of_faces] number of faces detected in face_img (0 in case of no faces detected)

                if multiple faces in the test image (face_img) the function will verfiy if the face provided in g_image appears in the test image or not.
        """
        
        no_of_faces, faces_coors = self.faces_locations(face_img)

        no_of_faces_gt,  gt_coors = None, None
        gt = None
        if not only_face_gt:
            no_of_faces_gt,  gt_coors = self.faces_locations(gt_img)
            if no_of_faces_gt != 1:
                raise Exception("Detected more than one face in the ground truth image ... ")
            gt   = self.verification_preprocess(gt_img,gt_coors[0])
            
        else:
            gt   = self.verification_preprocess(gt_img)



        if no_of_faces == 0:
            return False, no_of_faces
    
        distances = list()        
        
        for face_coor in faces_coors:
            face = self.verification_preprocess(face_img, face_coor)


            distance = self.face_similarity(face,gt)
            distances.append(distance)

        min_distance = min(distances)

        if min_distance < threshold:
            return True, no_of_faces

        return False, no_of_faces


    
    def get_faces(self, image, max_no_faces = 2):
        """
        input: BGR image
        return: list of faces as numpy array
        """
        _, faces_locations = self.faces_locations(image,max_no_faces)
        faces = list()
        for face_location in faces_locations:
            (x,y,w,h) = face_location
            face = image[y:y+h,x:x+w]
            face = cv2.cvtColor(face, cv2.COLOR_BGR2RGB)
            faces.append(face)

        return faces

    def faces_locations(self, image, max_no_faces = 2):
        """
            input: [image] frame given by cv.VideoCapture or image given from cv2.imread() (BGR images)
                   [max_no_faces] an integer represents the max. no. of faces to be detcted in the image 
            return: [int] no. of faces detected in images, 
                    [list] list of faces coordinates in shape of (x, y, w, h)
        
        """

        
        faces_coors = self.__faceDetector(image, max_no_faces)

        no_faces = len(faces_coors)

        return no_faces, faces_coors   
    
    def verification_preprocess(self,  img, coors = None):
        """
        input: coordinates given by face detection, image to extract the face, None if the image contains only one face
        operation: preprocessing the image to extract the face frome image and do some preprocessing for
        the faceEmbeddings model

        return: preprocessed image
        """
        img = img
        if coors is not None:
            (x,y,w,h) = coors
            img = img[y:y+h,x:x+w]
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, (160,160), interpolation=cv2.INTER_LINEAR)
        img = self.prewhiten(img)
        img = img.transpose([2, 0, 1])
        img = np.expand_dims(img, axis=0)

        return img

    def prewhiten(self, x):
        """
        Normalizing the vector
        """
        mean = np.mean(x)
        std = np.std(x)
        std_adj = np.maximum(std, 1.0 / np.sqrt(x.size))
        y = np.multiply(np.subtract(x, mean), 1 / std_adj)
        return y
    
    def face_similarity(self, face, gt):
        """
        input: face to verify [output from face_detection],
                ground_truth image
        output: verfied (boolen), True verified, False unverified

        """

        face_embeddings = self.face_embeddings(face)
        gt_embeddings = self.face_embeddings(gt)

        euclidean_distance = self.findEuclideanDistance(self.l2_normalize(face_embeddings), self.l2_normalize(gt_embeddings))

        
        return euclidean_distance


    

    def face_embeddings(self, face):
        """
        input: image of face with dimensions of 96x96
        return: face embeddings of the image
        """
        self.__faceEmbeddingNet.setInput(face)
        faceEmbeddings = self.__faceEmbeddingNet.forward()[0]
        return faceEmbeddings
    
    
    def findEuclideanDistance(self, source_embedding, test_embedding):
        """
        inputs: emebddings of two faces
        returns: euclidean distance between two embeddings (vectors)
        """
        euclidean_distance = source_embedding - test_embedding
        euclidean_distance = np.sum(np.multiply(euclidean_distance, euclidean_distance))
        euclidean_distance = np.sqrt(euclidean_distance)
        return euclidean_distance
    
    def l2_normalize(self, x):
        return x / np.sqrt(np.sum(np.multiply(x, x)))



import time
import cv2
import numpy as np
import onnxruntime
from blazeFaceUtils import gen_anchors, AnchorsOptions
import json
import time
import os

class BlazeFaceDetector:

    def __init__(self, typ = "front", scoreThreshold = 0.7, iouThreshold = 0.3):
        BASE_DIR = os.path.dirname(__file__)
        self.type = typ
        self.scoreThreshold = scoreThreshold
        self.iouThreshold = iouThreshold
        self.sigmoidScoreThreshold = np.log(self.scoreThreshold/(1-self.scoreThreshold))
        self.initModel(typ,BASE_DIR)

        self.generateAnchorBoxes(typ)
    
    
    def initModel(self, typ, BASE_DIR):
        """
        Read and initialize the model
        """
        front_model_path = os.path.join(BASE_DIR , "face_detection_front_128x128.onnx")
        back_model_path = os.path.join(BASE_DIR , "face_detection_back_256x256.onnx")
        if typ == "front":
            self.session = onnxruntime.InferenceSession(front_model_path, None)
        elif typ =="back":
            self.session = onnxruntime.InferenceSession(back_model_path, None)
       
        self.getModelInputDetails()
        self.getModelOutputDetails()
    
    def detectFaces(self, image, maxFaces = 2):
        """
        This function aims to detect faces in image
        input: [image] a numpy cv2 array (i.e. BGR image)
        returns:
        """
        preprocessed_image = self.detectionPreprocessing(image)
        result = self.inference(preprocessed_image)

        scores, goodDetections = self.filterDetections(result[1])

        faces_locations = self.extractDetections(result[0], goodDetections)
        faces_locations = self.nms(faces_locations, scores, maxFaces)
        return faces_locations
    
    def __call__(self, image, maxFaces=2):
        return self.detectFaces(image, maxFaces)

    
    
    def getModelInputDetails(self):
        """
        This function aims to get inlut node details for onnx inference
        """
        self.input_details = self.session.get_inputs()[0]
        self.input_name = self.input_details.name
        input_shape = self.input_details.shape
        self.inputHeight = input_shape[1]
        self.inputWidth = input_shape[2]
        self.channels = input_shape[3]

    def getModelOutputDetails(self):
        """
        this funtion aims to get the output nodes details (mainly we need the names only) for onnx inference
        """
        self.output_details = self.session.get_outputs()
    
    def generateAnchorBoxes(self, typ):
        """
        This function aims to generate anchor boxes depends on the type of the model (front model or back model)

        """
        if typ == "front":
		# Options to generate anchors for SSD object detection models.
            anchors_options = AnchorsOptions(input_size_width=128, input_size_height=128, min_scale=0.1484375, max_scale=0.75
                , anchor_offset_x=0.5, anchor_offset_y=0.5, num_layers=4
                , feature_map_width=[], feature_map_height=[]
                , strides=[8, 16, 16, 16], aspect_ratios=[1.0]
                , reduce_boxes_in_lowest_layer=False, interpolated_scale_aspect_ratio=1.0
                , fixed_anchor_size=True)

        elif typ == "back":
            # Options to generate anchors for SSD object detection models.
            anchors_options = AnchorsOptions(input_size_width=256, input_size_height=256, min_scale=0.15625, max_scale=0.75
                    , anchor_offset_x=0.5, anchor_offset_y=0.5, num_layers=4
                    , feature_map_width=[], feature_map_height=[]
                    , strides=[16, 32, 32, 32], aspect_ratios=[1.0]
                    , reduce_boxes_in_lowest_layer=False, interpolated_scale_aspect_ratio=1.0
                    , fixed_anchor_size=True)
        self.anchors = gen_anchors(anchors_options)

    def detectionPreprocessing(self, image):
        """
        This function aims to preproccess the image before feeding it to the onnx model
        input: [image] image loaded with cv2.imread (i.e. BGR image)
        returns: [data] processed image for onnx model feeding
        """
        img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        self.img_height, self.img_width, self.img_channels = img.shape
        # Input values should be from -1 to 1 with a size of 128 x 128 pixels for the fornt model
        # and 256 x 256 pixels for the back model
        img = img / 255.0
        """ img_resized = tf.image.resize(img, [self.inputHeight,self.inputWidth], 
                                    method='bicubic', preserve_aspect_ratio=False)
         """
        img_resized = cv2.resize(img, (self.inputHeight,self.inputWidth), interpolation=cv2.INTER_CUBIC)
        img_input = (img_resized - 0.5) / 0.5

        # Adjust matrix dimenstions
        reshape_img = img_input.reshape(1,self.inputHeight,self.inputWidth,self.channels)

        data = reshape_img.astype('float32')

        return data
    
    def inference(self, data):
        """
        This function takes the data from detectionPreprocessing function and passes it to BlazeFace model using onnx
        input: [data] data we got from detectionPreprocessing which is image as a numpy array
        return: [boxes] bounding boxes (including face landmarks)
                [scores] confidence of each output [boxes]
        """

        result = self.session.run([self.output_details[0].name,self.output_details[1].name,self.output_details[2].name,self.output_details[3].name], {self.input_name: data})
       
        scores = np.squeeze(np.concatenate((result[0], result[1]), axis = 1), axis =0)
        boxes = np.squeeze(np.concatenate((result[2], result[3]), axis = 1), axis = 0)
        
        return boxes, scores

    def filterDetections(self, scores):
        """
        This function aims to filter the bounding boxes by there confidence level
        input: [scores] numpy array with shape (num_boxes, 1) holds the confidence level of each box
        returns: numpy array scores after being filtered, and indices of those scores
        """
		# Filter based on the score threshold before applying sigmoid function
        goodDetections = np.where(scores > self.sigmoidScoreThreshold)[0]
        
		# Convert scores back from sigmoid values
        scores = 1.0 /(1.0 + np.exp(-scores[goodDetections]))
        
        return scores, goodDetections
    
    def extractDetections(self, predections, selectedIndices):
        """
        This function aims to extract the outputs from the anchor boxes
        inputs: [predections] numpy array holds the predicted outputs (boxes and keypoints), but we will use here the boxes only shape (num_of_boxes, 16)
                [selectedIndices] numpy array holds the indices of selected boxes
        returns: numpy array of boxes extracted from generated anchors shape (num_of_boxes, 4)
        """

        numGoodDetections = selectedIndices.shape[0]

        boxes = np.zeros((numGoodDetections, 4))
        for idx, detectionIdx in enumerate(selectedIndices):
            anchor = self.anchors[detectionIdx]

            sx = predections[detectionIdx, 0]
            sy = predections[detectionIdx, 1]
            w = predections[detectionIdx, 2]
            h = predections[detectionIdx, 3]

            cx = sx + anchor.x_center * self.inputWidth
            cy = sy + anchor.y_center * self.inputHeight

            cx /= self.inputWidth
            cy /= self.inputHeight
            w /= self.inputWidth
            h /= self.inputHeight
            
            # The next section is for facial keypoints detected from BlazeFace model 
            # commented it for speed concerns

            """ for j in range(KEY_POINT_SIZE):
                lx = output0[detectionIdx, 4 + (2 * j) + 0]
                ly = output0[detectionIdx, 4 + (2 * j) + 1]
                lx += anchor.x_center * self.inputWidth
                ly += anchor.y_center * self.inputHeight
                lx /= self.inputWidth
                ly /= self.inputHeight
                keypoints[idx,j,:] = np.array([lx, ly])
 """
            boxes[idx,:] = np.array([cx - w * 0.5, cy - h * 0.5, cx + w * 0.5, cy + h * 0.5])

        return boxes

    def nms(self, boxes, probs=None, max_num_outputs = 1, iouThreshold=0.3):
        """
        This a fast non max suppression (NMS) implementation
        inputs: [boxes] numpy array conatins bounding boxes in format of [x1, y1, x2, y2] with shape (num_boxes, 4)
                , [probs] numpy array holds the confidence level of each box from 0-1 with shape (num_boxes, 1)
                , [max_num_outputs] an intger holds the maximum number of ouputs needed
                , [iouThrshold] an integer from 0-1 represents the intersection over union threshold
        returns: numpy array of boxes after applying NMS
        """
	# if there are no boxes, return an empty list
        if len(boxes) == 0:
            return []

        # if the bounding boxes are integers, convert them to floats -- this
        # is important since we'll be doing a bunch of divisions
        if boxes.dtype.kind == "i":
            boxes = boxes.astype("float")

        # initialize the list of picked indexes
        pick = []

        # grab the coordinates of the bounding boxes
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]

        # compute the area of the bounding boxes and grab the indexes to sort
        # (in the case that no probabilities are provided, simply sort on the
        # bottom-left y-coordinate)
        area = (x2 - x1 + 1) * (y2 - y1 + 1)
        idxs = y2

        # if probabilities are provided, sort on them instead
        if probs is not None:
            idxs = probs.squeeze()

        # sort the indexes
        idxs = np.argsort(idxs)
        face_counter = 0
        # keep looping while some indexes still remain in the indexes list
        while len(idxs) > 0:
            # grab the last index in the indexes list and add the index value
            # to the list of picked indexes
            if face_counter == max_num_outputs:
                break
            face_counter +=1
            last = len(idxs) - 1
            i = idxs[last]
            
            pick.append(i)

            # find the largest (x, y) coordinates for the start of the bounding
            # box and the smallest (x, y) coordinates for the end of the bounding
            # box
            xx1 = np.maximum(x1[i], x1[idxs[:last]])
            yy1 = np.maximum(y1[i], y1[idxs[:last]])
            xx2 = np.minimum(x2[i], x2[idxs[:last]])
            yy2 = np.minimum(y2[i], y2[idxs[:last]])

            # compute the width and height of the bounding box
            w = np.maximum(0, xx2 - xx1 + 1)
            h = np.maximum(0, yy2 - yy1 + 1)

            # compute the ratio of overlap
            overlap = (w * h) / area[idxs[:last]]

            # delete all indexes from the index list that have overlap greater
            # than the provided overlap threshold
            idxs = np.delete(idxs, np.concatenate(([last],
                np.where(overlap > iouThreshold)[0])))
        # return only the bounding boxes that were picked
        
        boxes = boxes[pick]
        boxes[:,0] = self.img_width * boxes[:,0]
        boxes[:,2] = self.img_width * boxes[:,2]
        boxes[:,1] = self.img_height * boxes[:,1]
        boxes[:,3] = self.img_height * boxes[:,3]

        # convert from X1y1x2y2 to xywh
        boxes[:,2] = boxes[:,2] - boxes[:,0]
        boxes[:,3] = boxes[:,3] - boxes[:,1]
    
        return boxes.astype("int")
